"""
Parser for Harte chord data
"""

import os
from typing import Dict, List

import more_itertools as mitertools
from lark import Lark, Transformer

GRAMMAR = os.path.join(os.path.dirname(__file__), 'harte.lark')

with open(GRAMMAR, 'r', encoding='utf-8') as g:
    HARTE_LARK_GRAMMAR = g.read()


class TreeToHarteTransformer(Transformer):
    """
    Lark transformer to turn a parse tree into a Harte chord representation.
    The representation consists of a dict with keys:
      * root
          Root note of the chord
      * shorthand - OPTIONAL
          Shorthand of the chord
      * bass - OPTIONAL
          Modified bass note if slash chord is used
      * degrees - OPTIONAL
          Modified degrees on the chord (with missing degrees identified with *
           i.e. *3
    """
    @staticmethod
    def shorthand(shorthand: str) -> Dict[str, str]:
        """
        Extract the shorthand from the parse tree
        :param shorthand: shorthand of the chord
        :type shorthand: str
        :return: shorthand of the chord
        :rtype: str
        """
        return {"shorthand": str(shorthand[0])}

    @staticmethod
    def degree(degree: List[str]) -> str:
        """
        Extract the degree from the parse tree
        :param degree: degree of the chord
        :type degree: str
        :return: degree of the chord
        :rtype: str
        """
        return ''.join(degree)

    @staticmethod
    def bass(bass: List[str]) -> Dict[str, str]:
        """
        Extract the bass from the parse tree
        :param bass: bass of the chord
        :type bass: str
        :return: bass of the chord
        :rtype: str
        """
        return {'bass': ''.join(bass)}

    @staticmethod
    def note(root: List[str]) -> Dict[str, str]:
        """
        Extract the root from the parse tree
        :param root: root of the chord
        :type root: str
        :return: root of the chord
        :rtype: str
        """
        return {'root': ''.join(root)}

    @staticmethod
    def degree_list(degrees: List[str]) -> List[str]:
        """
        Extract the degrees from the parse tree
        :param degrees: degrees of the chord
        :type degrees: list
        :return: degrees of the chord
        :rtype: list
        """
        return degrees

    @staticmethod
    def chord(elements: List) -> Dict:
        """
        Method to transform a chord parse tree into a Harte chord representation
        :param elements: the chord parse tree
        :type elements: list
        :return: a Harte chord representation
        :rtype: dict
        """
        chord_dict = {}
        for elem in elements:
            if isinstance(elem, dict):
                chord_dict.update(**elem)
            elif isinstance(elem, list):
                chord_dict.update({"degrees": list(mitertools.collapse(elem))})

        return chord_dict


PARSER = Lark(HARTE_LARK_GRAMMAR,
              parser='lalr',
              start="chord",
              propagate_positions=False,
              maybe_placeholders=False,
              transformer=TreeToHarteTransformer())

if __name__ == '__main__':
    # test the grammar parsed Tree
    print(PARSER.parse('C:maj7(4,b6)/b4'))



"""
Extension of the Chord class from music21.chord to support the
Harte notation.
"""
# pylint: disable=consider-using-dict-items

from typing import List, Union

from music21.chord import Chord, ChordException
from music21.note import Note

from harte.interval import HarteInterval
from harte.mappings import SHORTHAND_DEGREES, DEGREE_SHORTHAND_MAP
from harte.parse_harte import PARSER
from harte.utils import degree_to_sort_key


class Harte(Chord):
    """
    Extension of the Chord class from music21.chord to support the
    Harte notation.
    """

    def __init__(self, chord: str, **keywords):
        """
        Constructor for the Harte class. It takes a Harte chord as input
        and parses it to extract the root, bass, degrees and shorthand. The
        constructor of the Chord class is then called to create the chord
        object
        :param chord: a music chord annotated according to the Harte notation
        :type chord: str
        """
        self.chord = chord
        try:
            parsed_chord = PARSER.parse(chord)
        except NameError as name_error:
            raise ChordException(
                f'The input chord {chord} is not a valid Harte chord') \
                from name_error

        if self.chord == 'N':
            self._root = self._bass = self._degrees = self._shorthand = None
            self._all_degrees = []
            super().__init__(self._all_degrees)
            return

        if "root" in parsed_chord:
            # chord is not empty
            # retrieve information from the parsed chord
            self._root = parsed_chord['root']
            self._shorthand = parsed_chord[
                'shorthand'] if 'shorthand' in parsed_chord.keys() else None
            self._degrees = parsed_chord[
                'degrees'] if 'degrees' in parsed_chord.keys() else None
            self._bass = parsed_chord[
                'bass'] if 'bass' in parsed_chord.keys() else '1'
            removed_degrees = [x for x in self._degrees if x.startswith(
                '*')] if self._degrees else []

            # unwrap shorthand if it exists and merge with degrees
            # if no shorthand exists, just use the degrees
            # if no degrees exist, assume the chord is a major triad
            if self._shorthand:
                assert SHORTHAND_DEGREES[self._shorthand], 'The Harte ' \
                                                           'shorthand is' \
                                                           ' not valid. '
                self._shorthand_degrees = SHORTHAND_DEGREES[self._shorthand]
                self._all_degrees = self._shorthand_degrees + self._degrees if \
                    self._degrees else self._shorthand_degrees
            elif self._degrees:
                self._all_degrees = self._degrees
            else:
                self._all_degrees = ['1', '3', '5']

            self._all_degrees = [x for x in self._all_degrees if
                                 x not in removed_degrees]
            # add root and bass note to the overall list of degrees
            self._all_degrees.append(self._bass)
            self._all_degrees.append('1')
            # sort the list and remove duplicates
            self._all_degrees = list(set(self._all_degrees))
            self._all_degrees.sort(key=degree_to_sort_key)

            # convert notes and interval to m21 primitives
            # note that when multiple flats are introduced (i.e. Cbb) music21
            # won't be able to parse the note.
            # this is fixed by replacing each 'b' with a '-'.
            m21_root = Note(self._root.replace("b", "-"))
            m21_degrees = [HarteInterval(x).transposeNote(m21_root)
                           for x in self._all_degrees]
            m21_bass = HarteInterval(self._bass).transposeNote(
                m21_root)
            # make sure that order between notes in the chord is respected
            # by transposing pitches higher than the preceding one when needed
            for i in range(1, len(m21_degrees)):
                m21_degrees[i].pitch.octave = m21_degrees[
                    i].pitch.implicitOctave
                m21_degrees[i].pitch.transposeAboveTarget(
                    m21_degrees[i - 1].pitch,
                    inPlace=True)
            m21_bass.pitch.octave = m21_bass.pitch.implicitOctave
            if m21_bass != m21_degrees[0]:
                m21_bass.pitch.transposeBelowTarget(m21_degrees[0].pitch,
                                                    inPlace=True)

            # initialize the parent constructor
            super().__init__(m21_degrees, **keywords)
            super().root(m21_root)
            super().bass(m21_bass)
        else:
            # chord is empty
            self._root = None
            self._shorthand = None
            self._degrees = None
            self._bass = None
            super().__init__(**keywords)

    def __deepcopy__(self, *args, **kwargs):
        """
        Perform a deepcopy of this obect by creating a new identical
        object with the input chord used for this one.

        :return: A copy of the current object.
        :rtype: Harte
        """
        return Harte(self.chord)

    def get_degrees(self) -> List[str]:
        """
        Method to retrieve the degrees of the chord in Harte notation without
        considering the degrees associated to the shorthand
        :return: a list of strings representing the degrees of the chord in
        Harte notation (e.g. ['b3', '5', '7'])
        :rtype: list[str]
        """
        return self._degrees

    def get_root(self) -> str:
        """
        Method to retrieve the root of the chord in Harte notation expressed as
        a string representing the note name (e.g. 'C#')
        :return: the root of the chord expressed as a string representing the
        note name (e.g. 'C#')
        :rtype: str
        """
        return self._root

    def get_bass(self) -> str:
        """
        Method to retrieve the bass of the chord in Harte notation expressed as
        a string representing the note interval calculated from the root note
        (e.g. 'b3)
        :return: the bass of the chord expressed as a string representing the
        note interval calculated from the root note (e.g. 'b3)
        """
        return self._bass

    def bass_is_root(self) -> bool:
        """
        Method to check if the bass of the chord is the same as the root of the
        chord
        :return: True if the bass of the chord is the same as the root of the
        chord, False otherwise
        :rtype: bool
        """
        if self._bass == self._root:
            return True
        return False

    def contains_shorthand(self) -> bool:
        """
        Method to check if the chord contains a shorthand or not
        :return: True if the chord contains a shorthand notation, False
        otherwise
        :rtype: bool
        """
        if self._shorthand:
            return True
        return False

    def get_shorthand(self) -> Union[str, bool]:
        """
        Method to retrieve the shorthand of the chord in Harte notation
        :return: the shorthand of the chord in Harte notation (e.g. 'maj7')
        if it exists, None otherwise
        :rtype: str
        """
        return self._shorthand

    def prettify(self) -> str:
        """
        Method to prettify the chord in Harte notation. It decomposes the chord
        into its constituent parts and returns a string representing the chord
        in Harte notation summarising the constituent degrees in shorthands,
        if possible
        :return: a string representing the chord in Harte notation summarising
        the constituent degrees in shorthands, if possible. If it is not
        possible to summarise the chord in a shorthand, the chord is returned
        in its original form
        :rtype: str
        """
        separator, shorthand = None, None
        degrees = self._all_degrees
        if '1' in degrees:
            degrees.remove('1')
        for grades in DEGREE_SHORTHAND_MAP.keys():
            intersection = set(grades).intersection(degrees)
            if len(intersection) == len(grades):
                shorthand = DEGREE_SHORTHAND_MAP[grades]
                clean_harte_degrees = list(
                    set(degrees) - intersection)
                if 'sus' in shorthand and '*3' in clean_harte_degrees:
                    clean_harte_degrees.remove('*3')
                break
        if shorthand:
            clean_harte_degrees = f'({",".join(clean_harte_degrees)})' if len(
                clean_harte_degrees) > 0 else ''
            if len(shorthand) > 0 or len(clean_harte_degrees) > 0:
                separator = ':'
            return self._root + separator + shorthand + clean_harte_degrees
        return self.chord

    def unwrap_shorthand(self) -> Union[List[str], None]:
        """
        Method to retrieve the degrees of the chord in Harte notation, both
        those associated to the shorthand and those explicitly specified
        :return: a list of strings representing the degrees of the chord in
        Harte notation, both those associated to the shorthand and those
        explicitly specified.
        """
        if self._shorthand:
            return self._all_degrees
        if self._degrees:
            return self._degrees
        return None

    def __eq__(self, other):
        """
        Method to check if two HarteChord objects are equal
        :param other: the other HarteChord object to compare to the current
        object
        :return: True if the two HarteChord objects are equal, False otherwise
        """
        if isinstance(other, Harte):
            return self._root == other.get_root() and self._degrees == \
                other.get_degrees() and self._bass == other.get_bass()
        return False

    def __repr__(self):
        """
        Method to represent the HarteChord object as a string
        :return: a string representing the HarteChord object
        """
        return f'Harte({str(self)})'

    def __str__(self):
        """
        Method to represent the HarteChord object as a string
        :return: a string representing the HarteChord object
        """
        if self._root is not None:
            chord_str = self._root
            if self._shorthand is not None:
                chord_str += ':' + self._shorthand
            if self._degrees is not None:
                chord_str += '(' + self._degrees + ')'
            if self._bass != '1':
                chord_str += '/' + self._bass
        else:
            chord_str = 'N'

        return chord_str


if __name__ == '__main__':
    # test utilities
    test_chord = Harte('N')
    root = test_chord.get_root()
    print(test_chord.pitchNames)
    print(test_chord.fullName)
    print(test_chord.commonName)
    print(test_chord.inversion())
    print(root, test_chord.bass(), test_chord.get_bass())
    print(test_chord.prettify())



"""
Extension of the Interval class from music21.interval to support the
Harte notation.
"""

from music21.interval import Interval, IntervalException
from music21.note import Note

from harte.utils import convert_interval


class HarteInterval(Interval):
    """
    music21 Interval class extension to support the Harte notation.
    """

    def __init__(self, harte_interval: str, **keywords) -> None:
        """
        Constructor for the HarteInterval class. It takes a string containing
        an interval as expressed by the Harte notation (e.g. 'b2') and aligns
        it to the music21 notation (e.g. 'm2'), by initializing the parent
        constructor
        :param harte_interval: A string containing an interval as expressed
        by the Harte notation (e.g. 'b2')
        :type harte_interval: str
        """
        self._harte_interval = harte_interval
        try:
            self._converted_interval = convert_interval(self._harte_interval)
        except ValueError as value_error:
            raise IntervalException(
                'Harte Interval cannot be converted') from value_error
        self._converted_interval = convert_interval(self._harte_interval)
        super().__init__(self._converted_interval, **keywords)

    def __eq__(self, other) -> bool:
        """
        Defines the equality operator for the HarteInterval class
        :param other: The other HarteInterval object to compare to
        :type other: HarteInterval
        :return: True if the two HarteInterval objects are equal, False
        otherwise.
        """
        if isinstance(other, HarteInterval):
            return self._harte_interval == other._harte_interval
        return self._harte_interval == other

    def __repr__(self) -> str:
        """
        Returns a string representation of the HarteInterval object
        :return: A string representation of the HarteInterval object
        """
        return f'HarteInterval({self._harte_interval})'

    def __str__(self) -> str:
        """
        Returns a string representation of the HarteInterval object
        :return: A string representation of the HarteInterval object
        """
        return self._harte_interval


if __name__ == '__main__':
    i = HarteInterval('3').transposeNote(Note('C'))
    print(i)



"""
Utility functions for processing Harte Chords
"""
# pylint: disable=consider-iterating-dictionary

import re
from typing import List

from harte.mappings import SHORTHAND_DEGREES


def convert_interval(harte_interval: str) -> str:
    """
    Utility function to convert a Harte interval to a music21 interval
    :param harte_interval: an interval of a Harte Chord
    :type harte_interval: str
    :return: the music21-shaped interval corresponding to the Harte interval
    format as a string
    :rtype: str
    """
    regex = r'([#]+)?([b]+)?(\d+)'
    matches = re.findall(regex, harte_interval)[0]

    base_degree = int(matches[2]) if int(matches[2]) < 8 else int(
        matches[2]) - 7

    if base_degree in [1, 4, 5]:
        if len(matches[0]) == 0 and len(matches[1]) == 0:
            modifier = 'P'
        elif len(matches[0]) == 1 and len(matches[1]) == 0:
            modifier = 'A'
        elif len(matches[0]) == 0 and len(matches[1]) == 1:
            modifier = 'd'
        elif len(matches[0]) >= 2 and len(matches[1]) == 0:
            new_sharps = matches[0].replace('##', '')
            return convert_interval(f'{new_sharps}{base_degree + 1}')
        elif len(matches[0]) == 0 and len(matches[1]) >= 2:
            new_flats = matches[1].replace('bb', '')
            return convert_interval(f'{new_flats}{base_degree - 1}')
        else:
            raise ValueError(f'The degree {harte_interval} cannot '
                             f'be parsed.')
    else:
        if len(matches[0]) == 0 and len(matches[1]) == 0:
            modifier = 'M'
        elif len(matches[0]) == 1 and len(matches[1]) == 0:
            modifier = 'A'
        elif len(matches[1]) == 1 and len(matches[0]) == 0:
            modifier = 'm'
        elif len(matches[1]) == 2 and len(matches[0]) == 0:
            modifier = 'd'
        else:
            raise ValueError(f'The degree {harte_interval} cannot '
                             f'be parsed.')

    return modifier + str(base_degree)


def unwrap_shorthand(harte_shorthand: str, harte_degrees: list) -> List[str]:
    """
    DEPRECATED
    Utility function to unwrap a shorthand notation into a list of degrees
    :param harte_shorthand: a shorthand notation of a Harte Chord
    :type harte_shorthand: str
    :param harte_degrees: a list of degrees of a Harte Chord
    :type harte_degrees: list
    :return: a list of degrees corresponding to the shorthand notation
    :rtype: List[str]
    """
    if harte_degrees or len(harte_degrees) > 0 and harte_shorthand != '':
        pass
    assert harte_shorthand in SHORTHAND_DEGREES.keys(), 'The Harte shorthand ' \
                                                        'is not valid. '

    shorthand_degrees = SHORTHAND_DEGREES[harte_shorthand]
    degrees = list(set(shorthand_degrees + harte_degrees))
    degrees.sort(key=lambda x: [k for k in x if k.isdigit()][0])
    return degrees


def degree_to_sort_key(degree: str) -> float:
    """
    Utility function for ordering degrees of a chord
    :param degree: a chord degree
    :type defree: str
    :return: a numerical value needed for ordering the degree
    :rtype: float
    """
    # extract number from degree
    degree_number = int("".join([k for k in degree if k.isdigit()]))

    # extract accidental from degree
    if degree.startswith('b'):
        degree_number -= 0.49
    elif degree.startswith('#'):
        degree_number += 0.49

    return degree_number


if __name__ == '__main__':
    print(unwrap_shorthand('minmaj9', ['b7', 'b9']))



"""
Setup file for the Harte Library package
"""

import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="harte-library",
    version="0.3.1",
    author="Andrea Poltronieri",
    description="Library for parsing Harte chords and converting them to Music21",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    py_modules=["harte"],
    package_dir={"harte-library": "harte"},
    install_requires=["music21", "numpy", "lark"],
    package_data={"": ["*.lark"]},
    include_package_data=True,
)



"""
Test cases for the harte module.
"""

import json
import os
from typing import List

import pytest

from harte.harte import Harte

# load a dict of chords frequencies extracted from ChoCo [1]
# to test coverage of a big set of chords
# [1] https://github.com/smashub/choco
CUR_DIR = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(CUR_DIR, "chords_count.json"), encoding="UTF-8") as f:
    CHORDS_COUNT = json.load(f)


@pytest.mark.parametrize("chord", list(CHORDS_COUNT.keys()))
def test_coverage(chord: str):
    """
    Tests coverage of all chord extracted from ChoCo [1].

    [1] https://github.com/smashub/choco

    :param chord: Chord to be tested
    :type chord: str
    """
    Harte(chord)


@pytest.mark.parametrize("chord,intervals",
                         [("C", ["P5", "M3"]),
                          ("A", ["P5", "M3"]),
                          ("C:maj", ["P5", "M3"]),
                          ("C:min", ["P5", "m3"]),
                          ("C:dim", ["d5", "m3"]),
                          ("C:aug", ["A5", "M3"]),
                          ("N", [])])
def test_interval_extraction(chord: str, intervals: List[str]):
    """
    Test that the annotateIntervals of music21 correctly works in extracting
    the intervals from a chord.

    :param chord: Input chord
    :type chord: str
    :param intervals: Intervals that should be part of the chord
    :type intervals: List[str]
    """
    chord = Harte(chord)
    annotated_intervals = chord.annotateIntervals(inPlace=False,
                                                  returnList=True,
                                                  stripSpecifiers=False)
    assert set(intervals) == set(annotated_intervals)


@pytest.mark.parametrize("chord,pitches", [("F:(b3, 5, b7, 11)",
                                            ["F", "A-", "C", "E-", "B-"]),
                                           ("F:(b3, 11, b7, 5)",
                                            ["F", "A-", "C", "E-", "B-"]),
                                           ("F:maj7(#11)",
                                            ["F", "A", "C", "E", "B"])])
def test_ordering_of_degrees(chord: str, pitches: List[str]):
    """
    Test that the parsed degrees are ordered correctly in the resulting m21 object.

    :param chord: Input chord
    :type chord: str
    :param pitches: Pitches that should be part of the chord
    :type pitches: List[str]
    """
    chord = Harte(chord)
    assert [p.name for p in chord.pitches] == pitches
